import logging
from importlib.metadata import version
from typing import Literal, NamedTuple
from .conf_file_utils import ConfFileUtils
from .crypto import Cryptography
from .exceptions import get_exception
from .file_utils import FileUtils
from .log import SizeRotatingLog, TimedRotatingLog
from .misc_utils import MiscUtils, Object
from .os_utils import OsUtils


__all__ = (
    "ConfFileUtils",
    "Cryptography",
    "get_exception",
    "FileUtils",
    "TimedRotatingLog",
    "SizeRotatingLog",
    "MiscUtils",
    "Object",
    "OsUtils"
)

__title__ = "ddcUtils"
__author__ = "Daniel Costa"
__email__ = "danieldcsta@gmail.com>"
__license__ = "MIT"
__copyright__ = "Copyright 2023-present ddc"
_req_python_version = (3, 11, 0)


try:
    _version = tuple(int(x) for x in version(__title__).split("."))
except ModuleNotFoundError:
    _version = (0, 0, 0)


class VersionInfo(NamedTuple):
    major: int
    minor: int
    micro: int
    releaselevel: Literal["alpha", "beta", "candidate", "final"]
    serial: int


__version__ = _version
__version_info__: VersionInfo = VersionInfo(
    major=__version__[0],
    minor=__version__[1],
    micro=__version__[2],
    releaselevel="final",
    serial=0
)
__req_python_version__: VersionInfo = VersionInfo(
    major=_req_python_version[0],
    minor=_req_python_version[1],
    micro=_req_python_version[2],
    releaselevel="final",
    serial=0
)

logging.getLogger(__name__).addHandler(logging.NullHandler())

del logging, NamedTuple, Literal, VersionInfo, version, _version, _req_python_version



# -*- encoding: utf-8 -*-
import base64
import os
from cryptography.fernet import Fernet, InvalidToken


class Cryptography:
    def __init__(self, private_key: str = None):
        self.private_key = private_key or "sMZo38VwRdigN78FBnHj8mETNlofL4Qhj_x5cvyxJsc="
        try:
            self.cipher_suite = Fernet(bytes(self.private_key, "UTF-8"))
        except ValueError as e:
            raise ValueError(str(e))

    @staticmethod
    def generate_private_key() -> str:
        """
        Generates a private key to be used instead of default one
        But keep in mind that this private key will be needed to decode further strings
        :return: str
        """

        private_key = base64.urlsafe_b64encode(os.urandom(32))
        return private_key.decode("UTF-8")

    def encode(self, str_to_encode: str) -> str:
        """
        Encodes a given string
        :param str_to_encode:
        :return: str
        """

        str_bytes = bytes(str_to_encode, "UTF-8")
        encoded_text = self.cipher_suite.encrypt(str_bytes)
        return encoded_text.decode("UTF-8")

    def decode(self, str_to_decode: str) -> str:
        """
        Decodes a given string
        :param str_to_decode:
        :return:
        """

        if str_to_decode is not None and len(str_to_decode) > 0:
            try:
                bet = bytes(str_to_decode, "UTF-8")
                decoded_text = self.cipher_suite.decrypt(bet).decode("UTF-8")
                return str(decoded_text)
            except InvalidToken:
                error_msg = "Not encrypted"
                if len(str_to_decode) == 100:
                    error_msg = "Encrypted with another private key"
                raise InvalidToken(error_msg)



# -*- encoding: utf-8 -*-
import errno
import gzip
import json
import os
import shutil
import struct
import subprocess
import sys
from datetime import datetime, timedelta
from pathlib import Path
from zipfile import ZipFile
import fsspec
import requests
from .exceptions import get_exception
from .os_utils import OsUtils


class FileUtils:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    @staticmethod
    def show(path: str) -> bool:
        """
        Open the given file or directory in explorer or notepad and returns True for success or False for failed access
        :param path:
        :return: bool
        """

        if not os.path.exists(path):
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), path)
        try:
            return_code = 0
            match OsUtils.get_os_name():
                case "Windows":
                    os.startfile(path)
                case "Darwin":
                    return_code = subprocess.call(("open", path))
                case _:
                    return_code = subprocess.call(("xdg-open", path))
            return not bool(return_code)
        except Exception as e:
            sys.stderr.write(get_exception(e))
            raise e

    @staticmethod
    def list_files(directory: str, starts_with: str = None, ends_with: str = None) -> list:
        """
        List all files in the given directory and returns them in a list sorted by creation time in ascending order
        :param directory:
        :param starts_with:
        :param ends_with:
        :return: list
        """

        try:
            result_list = []
            if os.path.isdir(directory):
                if starts_with and ends_with:
                    result_list = [Path(os.path.join(directory, f)) for f in os.listdir(directory) if
                                   f.lower().startswith(starts_with.lower()) and
                                   f.lower().endswith(ends_with.lower())]
                elif starts_with:
                    result_list = [Path(os.path.join(directory, f)) for f in os.listdir(directory) if
                                   f.lower().startswith(starts_with.lower())]
                elif ends_with:
                    result_list = [Path(os.path.join(directory, f)) for f in os.listdir(directory) if
                                   f.lower().endswith(ends_with.lower())]
                else:
                    result_list = [Path(os.path.join(directory, f)) for f in os.listdir(directory)]
                result_list.sort(key=os.path.getctime)
            return result_list
        except Exception as e:
            sys.stderr.write(get_exception(e))
            raise e

    @staticmethod
    def gzip(input_file_path: str, output_dir: str = None) -> Path | None:
        """
        Compress the given file and returns the Path for success or None if failed
        :param input_file_path:
        :param output_dir:
        :return: Path | None:
        """

        if not output_dir:
            output_dir = os.path.dirname(input_file_path)

        input_file_name = os.path.basename(input_file_path)
        output_filename = f"{os.path.splitext(input_file_name)[0]}.gz"
        output_file = os.path.join(output_dir, output_filename)

        try:
            with open(input_file_path, "rb") as fin:
                with gzip.open(output_file, "wb") as fout:
                    fout.writelines(fin)
            return Path(output_file)
        except Exception as e:
            sys.stderr.write(get_exception(e))
            if os.path.isfile(output_file):
                os.remove(output_file)
            raise e

    @staticmethod
    def unzip(file_path: str, out_path: str = None) -> ZipFile | None:
        """
        Unzips the given file.zip and returns ZipFile for success or None if failed
        :param file_path:
        :param out_path:
        :return: ZipFile | None
        """

        try:
            out_path = out_path or os.path.dirname(file_path)
            with ZipFile(file_path) as zipf:
                zipf.extractall(out_path)
            return zipf
        except Exception as e:
            sys.stderr.write(get_exception(e))
            raise e

    @staticmethod
    def remove(path: str) -> bool:
        """
        Remove the given file and returns True if the file was successfully removed
        :param path:
        :return:
        """
        try:
            if os.path.isfile(path):
                os.remove(path)
            elif os.path.exists(path):
                shutil.rmtree(path)
        except OSError as e:
            sys.stderr.write(get_exception(e))
            raise e
        return True

    @staticmethod
    def rename(from_name: str, to_name: str) -> bool:
        """
        Rename the given file and returns True if the file was successfully
        :param from_name:
        :param to_name:
        :return: bool
        """

        try:
            if os.path.exists(from_name):
                os.rename(from_name, to_name)
        except OSError as e:
            sys.stderr.write(get_exception(e))
            raise e
        return True

    @staticmethod
    def copy_dir(src, dst, symlinks=False, ignore=None) -> bool:
        """
        Copy files from src to dst and returns True or False
        :param src:
        :param dst:
        :param symlinks:
        :param ignore:
        :return: True or False
        """

        try:
            for item in os.listdir(src):
                s = os.path.join(src, item)
                d = os.path.join(dst, item)
                if os.path.isdir(s):
                    shutil.copytree(s, d, symlinks, ignore)
                else:
                    shutil.copy2(s, d)
        except IOError as e:
            sys.stderr.write(get_exception(e))
            return False
        return True

    @staticmethod
    def download_file(remote_file_url, local_file_path) -> bool:
        """
        Download file from remote url to local and returns True or False
        :param remote_file_url:
        :param local_file_path:
        :return: True or False
        """

        try:
            req = requests.get(remote_file_url)
            if req.status_code == 200:
                with open(local_file_path, "wb") as outfile:
                    outfile.write(req.content)
                return True
        except requests.HTTPError as e:
            sys.stderr.write(get_exception(e))
        return False

    def download_github_dir(self, remote_dir_url: str, local_dir_path: str) -> bool:
        """
        Download directory from remote url to local and returns True or False
        Need to specify the branch on remote url
            example: https://github.com/ddc/ddcutils/blob/master/ddcutils/databases

        :param remote_dir_url:
        :param local_dir_path:
        :return:
        """

        try:
            if not os.path.exists(local_dir_path):
                os.makedirs(local_dir_path, exist_ok=True)

            req_dir = requests.get(remote_dir_url)
            if req_dir.status_code == 200:
                data_dict = json.loads(req_dir.content)
                files_list = data_dict["payload"]["tree"]["items"]
                for file in files_list:
                    remote_file_url = f"{remote_dir_url}/{file['name']}"
                    local_file_path = f"{local_dir_path}/{file['name']}"
                    if file["contentType"] == "directory":
                        self.download_github_dir(remote_file_url, local_file_path)
                    else:
                        req_file = requests.get(remote_file_url)
                        if req_file.status_code == 200:
                            data_dict = json.loads(req_file.content)
                            content = data_dict["payload"]["blob"]["rawLines"]
                            if not content:
                                payload = data_dict['payload']
                                url = (f"https://raw.githubusercontent.com/"
                                       f"{payload['repo']['ownerLogin']}/"
                                       f"{payload['repo']['name']}/"
                                       "master/"
                                       f"{payload['path']}")
                                req_file = requests.get(url)
                                with open(local_file_path, "wb") as outfile:
                                    outfile.write(req_file.content)
                            else:
                                with open(local_file_path, "w") as outfile:
                                    outfile.writelines([f"{line}\n" for line in content])

        except Exception as e:
            sys.stderr.write(get_exception(e))
            return False
        return True

    @staticmethod
    def get_exe_binary_type(file_path: str) -> str | None:
        """
        Returns the binary type of the given EXE file
        :param file_path:
        :return: str | None
        """

        with open(file_path, "rb") as f:
            s = f.read(2)
            if s != b"MZ":
                return "Not an EXE file"
            f.seek(60)
            s = f.read(4)
            header_offset = struct.unpack("<L", s)[0]
            f.seek(header_offset + 4)
            s = f.read(2)
            machine = struct.unpack("<H", s)[0]
            match machine:
                case 332:
                    # IA32 (32-bit x86)
                    binary_type = "IA32"
                case 512:
                    # IA64 (Itanium)
                    binary_type = "IA64"
                case 34404:
                    # IAMD64 (64-bit x86)
                    binary_type = "AMD64"
                case 452:
                    # IARM eabi (32-bit)
                    binary_type = "ARM-32bits"
                case 43620:
                    # IAArch64 (ARM-64, 64-bit)
                    binary_type = "ARM-64bits"
                case _:
                    binary_type = None
        return binary_type

    @staticmethod
    def is_older_than_x_days(path: str, days: int) -> bool:
        """
        Check if a file or directory is older than the specified number of days
        :param path:
        :param days:
        :return:
        """

        if not os.path.exists(path):
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), path)

        if int(days) == 1:
            cutoff_time = datetime.today()
        else:
            cutoff_time = datetime.today() - timedelta(days=int(days))

        stats = os.stat(path)
        days_epoch = cutoff_time.timestamp()
        if stats.st_ctime < days_epoch:
            return True
        return False

    @staticmethod
    def copy(src_path, dst_path):
        """
        Copy a file to another location
        :param src_path:
        :param dst_path:
        :return:
        """

        try:

            shutil.copy(src_path, dst_path)
        except Exception as e:
            return e
        return True

    @staticmethod
    def download_filesystem_directory(org: str,
                                      repo: str,
                                      branch: str,
                                      remote_dir: str,
                                      local_dir: str,
                                      filesystem: str = "github",
                                      exist_ok: bool = True,
                                      parents: bool = True,
                                      recursive: bool = False) -> bool:
        """
        Downloads a filesystem directory and save it to a local directory
        :param org:
        :param repo:
        :param branch:
        :param remote_dir:
        :param local_dir:
        :param filesystem:
        :param exist_ok:
        :param parents:
        :param recursive:
        :return:
        """

        try:
            destination = Path(local_dir)
            destination.mkdir(exist_ok=exist_ok, parents=parents)
            fs = fsspec.filesystem(filesystem, org=org, repo=repo, sha=branch)
            remote_files = fs.ls(remote_dir)
            fs.get(remote_files, destination.as_posix(), recursive=recursive)
        except requests.HTTPError as e:
            sys.stderr.write(get_exception(e))
            return False
        return True



# -*- encoding: utf-8 -*-
import sys
import sqlalchemy as sa
from sqlalchemy.engine import create_engine, Engine
from sqlalchemy.orm import Session, sessionmaker
from ..exceptions import get_exception


class DBSqlite:
    """
    Class to handle sqlite databases

    database = DBSqlite(DATABASE_FILE_PATH)
    with database.session() as session:
        do your stuff here

    """

    def __init__(self, db_file_path: str, batch_size=100, echo=False, future=True):
        self.file = db_file_path
        self.batch_size = batch_size
        self.echo = echo
        self.future = future

    def url(self) -> str:
        return f"sqlite:///{self.file}"

    def engine(self) -> Engine | None:
        try:
            engine = create_engine(self.url(), future=self.future, echo=self.echo).\
                execution_options(stream_results=self.echo, isolation_level="AUTOCOMMIT")

            @sa.event.listens_for(engine, "before_cursor_execute")
            def receive_before_cursor_execute(conn,
                                              cursor,
                                              statement,
                                              params,
                                              context,
                                              executemany):
                cursor.arraysize = self.batch_size
            return engine
        except Exception as e:
            sys.stderr.write(f"Unable to Create Database Engine: {get_exception(e)}")
            return None

    def session(self, engine: Engine = None) -> Session | None:
        _engine = engine or self.engine()
        if _engine is None:
            sys.stderr.write("Unable to Create Database Session: Empty Engine")
            return None
        session_maker = sessionmaker(bind=_engine)
        _engine.dispose()
        return session_maker()



# -*- encoding: utf-8 -*-
import sys
import sqlalchemy as sa
from sqlalchemy.orm import Session, sessionmaker
from sqlalchemy.ext.asyncio import async_sessionmaker, AsyncEngine, AsyncSession, create_async_engine
from ..exceptions import get_exception


class DBPostgres:
    """
    Class to handle postgres database connection

    database = DBPostgres(**db_configs)
    with database.session() as session:
        do your stuff here

    """

    def __init__(self, future=True, echo=False, **kwargs):
        self.echo = echo
        self.future = future
        self.username = kwargs["username"]
        self.password = kwargs["password"]
        self.host = kwargs["host"]
        self.port = kwargs["port"]
        self.db = kwargs["database"]

    def uri(self, drivername="psycopg2") -> sa.engine.URL:
        credentials = {
            "drivername": f"postgresql+{drivername}",
            "username": self.username,
            "password": self.password,
            "host": self.host,
            "port": self.port,
            "database": self.db
        }
        return sa.engine.URL.create(**credentials)

    def engine(self) -> sa.Engine | None:
        try:
            engine = sa.create_engine(
                self.uri(),
                echo=self.echo,
                future=self.future
            )
            return engine
        except Exception as e:
            sys.stderr.write(f"Unable to Create Database Engine: {get_exception(e)}")
            return None

    def session(self, engine: sa.Engine = None) -> Session | None:
        _engine = engine or self.engine()
        if _engine is None:
            sys.stderr.write("Unable to Create Database Session: Empty Engine")
            return None
        session = sessionmaker(
            bind=_engine,
            autoflush=True,
            expire_on_commit=False,
            future=self.future,
            class_=Session
        )
        return session()


class DBPostgresAsync(DBPostgres):
    """
    Class to handle async postgres database connection

        database = DBPostgresAsync(**db_configs)
        async with database.session() as session:
            do your stuff here

    """

    def __init__(self, future=True, echo=False, **kwargs):
        super().__init__(future, echo, **kwargs)

    def url(self, drivername="asyncpg") -> sa.engine.URL:
        return super().uri(drivername)

    def engine(self) -> AsyncEngine | None:
        try:
            engine = create_async_engine(
                self.url(),
                echo=self.echo,
                future=self.future
            )
            return engine
        except Exception as e:
            sys.stderr.write(f"Unable to Create Database Engine: {get_exception(e)}")
            return None

    def session(self, engine: AsyncEngine = None) -> AsyncSession | None:
        _engine = engine or self.engine()
        if _engine is None:
            sys.stderr.write("Unable to Create Database Session: Empty Engine")
            return None
        async_session = async_sessionmaker(
            bind=_engine,
            autoflush=True,
            expire_on_commit=False,
            future=self.future,
            class_=AsyncSession
        )
        return async_session()



# -*- encoding: utf-8 -*-
import gzip
import logging.handlers
import os
import sys
from logging.handlers import RotatingFileHandler, TimedRotatingFileHandler
from .exceptions import get_exception
from .file_utils import FileUtils


class TimedRotatingLog:
    """
    TimedRotatingLog class

    Current 'when' events supported:
    S - Seconds
    M - Minutes
    H - Hours
    D - Days
    midnight - roll over at midnight
    W{0-6} - roll over on a certain day; 0 - Monday
    """

    def __init__(
        self,
        level: str = "info",
        directory: str = "logs",
        filename: str = "app.log",
        encoding: str = "UTF-8",
        days_to_keep: int = 7,
        when: str = "midnight",
        utc: bool = True,
    ):
        self.level = _get_level(level)
        self.directory = directory
        self.filename = filename
        self.encoding = encoding
        self.days_to_keep = days_to_keep
        self.when = when
        self.utc = utc

    def init(self):
        log_file_path = _get_log_path(self.directory, self.filename)
        file_hdlr = TimedRotatingFileHandler(filename=log_file_path,
                                             encoding=self.encoding,
                                             when=self.when,
                                             utc=self.utc,
                                             backupCount=self.days_to_keep)
        file_hdlr.suffix = "%Y%m%d"
        file_hdlr.rotator = GZipRotatorTimed(self.directory, self.days_to_keep)
        return _set_log_format(file_hdlr, self.level)


class SizeRotatingLog:
    """
    SizeRotatingLog class
    """

    def __init__(
        self,
        level: str = "info",
        directory: str = "logs",
        filename: str = "app.log",
        encoding: str = "UTF-8",
        days_to_keep: int = 7,
        max_mbytes: int = 5
    ):
        self.level = _get_level(level)
        self.directory = directory
        self.filename = filename
        self.encoding = encoding
        self.days_to_keep = days_to_keep
        self.max_mbytes = max_mbytes

    def init(self):
        log_file_path = _get_log_path(self.directory, self.filename)
        file_hdlr = RotatingFileHandler(filename=log_file_path,
                                        mode="a",
                                        maxBytes=self.max_mbytes * 1024 * 1024,
                                        backupCount=self.days_to_keep,
                                        encoding=self.encoding,
                                        delay=False,
                                        errors=None)
        file_hdlr.rotator = GZipRotatorSize(self.directory, self.days_to_keep)
        return _set_log_format(file_hdlr, self.level)


class GZipRotatorSize:
    def __init__(self, dir_logs: str, days_to_keep: int):
        self.dir = dir_logs
        self.days_to_keep = days_to_keep

    def __call__(self, source: str, dest: str) -> None:
        RemoveOldLogs(self.dir, self.days_to_keep)
        if os.path.isfile(source) and os.stat(source).st_size > 0:
            new_file_number = 1
            old_gz_files_list = FileUtils.list_files(self.dir, ends_with=".gz")
            if old_gz_files_list:
                try:
                    oldest_file_number = int(old_gz_files_list[-1].name.split(".")[0].split("_")[1])
                    new_file_number = oldest_file_number + 1
                except ValueError as e:
                    _write_stderr(f"[Unable to get old zip log file number]:{get_exception(e)}: {old_gz_files_list[-1].name}")
            _gzip_file(source, new_file_number)


class GZipRotatorTimed:
    def __init__(self, dir_logs: str, days_to_keep: int):
        self.dir = dir_logs
        self.days_to_keep = days_to_keep

    def __call__(self, source: str, dest: str) -> None:
        RemoveOldLogs(self.dir, self.days_to_keep)
        output_dated_name = os.path.splitext(dest)[1].replace(".", "")
        _gzip_file(source, output_dated_name)


class RemoveOldLogs:
    def __init__(self, logs_dir: str, days_to_keep: int) -> None:
        files_list = FileUtils.list_files(logs_dir, ends_with=".gz")
        for file in files_list:
            file_path = str(os.path.join(logs_dir, file))
            if FileUtils.is_older_than_x_days(file_path, days_to_keep):
                try:
                    FileUtils.remove(file_path)
                except Exception as e:
                    _write_stderr(f"[Unable to remove old logs]:{get_exception(e)}: {file_path}")


def _write_stderr(msg: str) -> None:
    sys.stdout.write(f"[ERROR]:{msg}\n")


def _get_level(level: str) -> logging:
    if not isinstance(level, str):
        _write_stderr("[Unable to get log level]. Default level to: 'info'")
        return logging.INFO
    match level.lower():
        case "debug":
            return logging.DEBUG
        case "warning":
            return logging.WARNING
        case "error":
            return logging.ERROR
        case "critical":
            return logging.CRITICAL
        case _:
            return logging.INFO


def _get_log_path(directory: str, filename: str) -> str:
    try:
        os.makedirs(directory, exist_ok=True) if not os.path.isdir(directory) else None
    except Exception as e:
        _write_stderr(f"[Unable to create logs directory]:{get_exception(e)}: {directory}")
        raise e

    log_file_path = str(os.path.join(directory, filename))

    try:
        open(log_file_path, "a+").close()
    except IOError as e:
        _write_stderr(f"[Unable to open log file for writing]:{get_exception(e)}: {log_file_path}")
        raise e

    return log_file_path


def _set_log_format(file_hdlr, level: logging) -> logging.Logger:
    _debug_formatt = ""
    if level == logging.DEBUG:
        _debug_formatt = f"[PID:{os.getpid()}]:[%(filename)s:%(funcName)s:%(lineno)d]:"

    formatt = f"[%(asctime)s.%(msecs)03d]:[%(levelname)s]:{_debug_formatt}%(message)s"
    formatter = logging.Formatter(formatt, datefmt="%Y-%m-%dT%H:%M:%S")

    logger = logging.getLogger()
    logger.setLevel(level)

    file_hdlr.setFormatter(formatter)
    file_hdlr.setLevel(level)
    logger.addHandler(file_hdlr)

    stream_hdlr = logging.StreamHandler()
    stream_hdlr.setFormatter(formatter)
    stream_hdlr.setLevel(level)
    logger.addHandler(stream_hdlr)

    return logger


def _gzip_file(source, output_partial_name):
    if os.path.isfile(source) and os.stat(source).st_size > 0:
        try:
            sfname, sext = os.path.splitext(source)
            renamed_dst = f"{sfname}_{output_partial_name}{sext}.gz"
            with open(source, "rb") as fin:
                with gzip.open(renamed_dst, "wb") as fout:
                    fout.writelines(fin)
            FileUtils.remove(source)
        except Exception as e:
            _write_stderr(f"[Unable to zip log file]:{get_exception(e)}: {source}")



# -*- encoding: utf-8 -*-
import configparser
import errno
import os
import sys
from .exceptions import get_exception


class ConfFileUtils:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    @staticmethod
    def _get_default_parser() -> configparser.ConfigParser:
        """
        Returns the parser
        :return configparser.ConfigParser:
        """

        parser = configparser.ConfigParser(delimiters="=", allow_no_value=True)
        parser.optionxform = str  # this will not change all values to lowercase
        parser._interpolation = configparser.ExtendedInterpolation()
        return parser

    @staticmethod
    def _get_parser_value(parser: configparser.ConfigParser, section: str, config_name: str) -> str | int | None:
        """
        Returns the value of the specified section in the given parser
        :param parser:
        :param section:
        :param config_name:
        :return: str | int | None
        """

        try:
            value = parser.get(section, config_name).replace("\"", "")
            lst_value = list(value.split(","))
            if len(lst_value) > 1:
                values = []
                for each in lst_value:
                    values.append(int(each.strip()) if each.strip().isnumeric() else each.strip())
                value = values
            elif value is not None and type(value) is str:
                if len(value) == 0:
                    value = None
                elif value.isnumeric():
                    value = int(value)
                elif "," in value:
                    value = sorted([x.strip() for x in value.split(",")])
            else:
                value = None
        except Exception as e:
            sys.stderr.write(get_exception(e))
            value = None
        return value

    def _get_section_data(self, parser: configparser.ConfigParser, section: str, final_data: dict, mixed_values: bool = True, include_section_name: bool = False):
        """
        Returns the section data from the given parser
        :param parser:
        :param section:
        :param final_data:
        :param mixed_values:
        :param include_section_name:
        :return: dict
        """

        for name in parser.options(section):
            section_name = section.replace(" ", "_")
            config_name = name.replace(" ", "_")
            value = self._get_parser_value(parser, section, name)
            if mixed_values and include_section_name:
                final_data[f"{section_name}.{config_name}"] = value
            elif mixed_values and not include_section_name:
                final_data[config_name] = value
            else:
                final_data[section_name][config_name] = value
        return final_data

    def get_all_values(self, file_path: str, mixed_values: bool = False) -> dict:
        """
        Get all values from an .ini config file structure and returns them as a dictionary
        :param file_path:
        :param mixed_values:
        :return: dict
        """

        if not os.path.isfile(file_path):
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), file_path)
        final_data = {}
        parser = self._get_default_parser()
        try:
            parser.read(file_path)
            for section in parser.sections():
                if not mixed_values:
                    section_name = section.replace(" ", "_")
                    final_data[section_name] = {}
                final_data = self._get_section_data(parser, section, final_data, mixed_values, True)
        except Exception as e:
            sys.stderr.write(get_exception(e))
        return final_data

    def get_section_values(self, file_path: str, section: str) -> dict:
        """
        Get all section values from an .ini config file structure and returns them as a dictionary
        :param file_path:
        :param section:
        :return: dict
        """

        if not os.path.isfile(file_path):
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), file_path)
        final_data = {}
        parser = self._get_default_parser()
        try:
            parser.read(file_path)
            final_data = self._get_section_data(parser, section, final_data)
        except Exception as e:
            sys.stderr.write(get_exception(e))
        return final_data

    def get_value(self, file_path: str, section: str, config_name: str) -> str | int | None:
        """
        Get value from an .ini config file structure and returns it
        :param file_path:
        :param section:
        :param config_name:
        :return: str | int | None
        """

        if not os.path.isfile(file_path):
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), file_path)
        parser = self._get_default_parser()
        parser.read(file_path)
        value = self._get_parser_value(parser, section, config_name)
        return value

    def set_value(self, file_path: str, section_name: str, config_name: str, new_value, commas: bool = False) -> bool:
        """
        Set value from an .ini config file structure and returns True or False
        :param file_path:
        :param section_name:
        :param config_name:
        :param new_value:
        :param commas:
        :return: True or False
        """

        if not os.path.isfile(file_path):
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), file_path)
        parser = self._get_default_parser()
        parser.read(file_path)
        if commas:
            new_value = f'"{new_value}"'
        parser.set(section_name, config_name, new_value)
        try:
            with open(file_path, "w") as configfile:
                parser.write(configfile, space_around_delimiters=False)
        except configparser.DuplicateOptionError as e:
            sys.stderr.write(get_exception(e))
            return False
        return True



# -*- encoding: utf-8 -*-
from datetime import datetime, timezone
import json
import os
from pathlib import Path
from .os_utils import OsUtils
from . import constants


class Object:
    """
    This class is used for creating a simple class object
    """

    def __init__(self):
        self._created = datetime.now().isoformat()

    def to_json(self):
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=4)

    def to_dict(self):
        return json.loads(self.to_json())


class MiscUtils:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    @staticmethod
    def clear_screen() -> None:
        """
        Clears the terminal screen
        :return:
        """

        cmd = "cls" if OsUtils.is_windows() else "clear"
        os.system(cmd)

    @staticmethod
    def user_choice() -> input:
        """
        This function will ask the user to select an option
        :return: input
        """

        try:
            return input(">>> ").lower().strip()
        except SyntaxError:
            pass

    @staticmethod
    def get_active_branch_name(git_dir: str, master_branch_name: str = "master") -> str | None:
        """
        Returns the name of the active branch if found, else returns the "master" branch
        :return: str
        """

        head_dir = Path(os.path.join(git_dir, "HEAD"))
        try:
            with head_dir.open("r") as f:
                content = f.read().splitlines()
            for line in content:
                if line[0:4] == "ref:":
                    return line.partition("refs/heads/")[2]
        except FileNotFoundError:
            return master_branch_name

    @staticmethod
    def get_current_date_time() -> datetime:
        """
        Returns the current date and time on UTC timezone
        :return: UTC datetime
        """

        return datetime.now(timezone.utc)

    @staticmethod
    def convert_datetime_to_str_long(date: datetime) -> str:
        """
        Converts a datetime object to a long string
        :param date:
        :return: str
        """

        return date.strftime(constants.DATE_TIME_FORMATTER_STR)

    @staticmethod
    def convert_datetime_to_str_short(date: datetime) -> str:
        """
        Converts a datetime object to a short string
        :param date:
        :return: str
        """

        return date.strftime(f"{constants.DATE_FORMATTER} {constants.TIME_FORMATTER}")

    @staticmethod
    def convert_str_to_datetime_short(datetime_str: str) -> datetime:
        """
        Converts a str to a datetime
        :param datetime_str:
        :return: datetime
        """

        return datetime.strptime(datetime_str, f"{constants.DATE_FORMATTER} {constants.TIME_FORMATTER}")

    def get_current_date_time_str_long(self) -> str:
        """
        Returns the current date and time as string
        :return: str
        """

        return self.convert_datetime_to_str_long(self.get_current_date_time())
